# Copyright (c) 2007, 2008 Red Hat, Inc.
#
# This file is part of the Qpid async store library msgstore.so.
#
# This library is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License as published by the Free Software Foundation; either
# version 2.1 of the License, or (at your option) any later version.
#
# This library is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public
# License along with this library; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301
# USA
#
# The GNU Lesser General Public License is available in the file COPYING.

import os.path, sys, xml.parsers.expat
from struct import pack, unpack, calcsize
from time import gmtime, strftime

# TODO: Get rid of these! Use jinf instance instead
dblkSize = 128
sblkSize = 4 * dblkSize

#== protected and private ======================================================================

_extern_mask    = 0x20

class Utils(object):

    __printchars = "0123456789abcdefghijklmnopqrstuvwxyzABCDEFGHIJKLMNOPQRSTUVWXYZ!\"#$%&'()*+,-./:;<=>?@[\\]^_`{\|}~ "
    
    # The @staticmethod declarations are not supported in RHEL4 (python 2.3.x)
    # When RHEL4 support ends, restore these declarations and remove the older
    # staticmethod() declaration.

    #@staticmethod
    def formatData(dsize, data):
        if data == None:
            return ""
        if Utils.__isPrintable(data):
            datastr = Utils.__splitStr(data)
        else:
            datastr = Utils.__hexSplitStr(data)
        if dsize != len(data):
            raise Exception("Inconsistent data size: dsize=%d, data(%d)=\"%s\"" % (dsize, len(data), datastr))
        return "data(%d)=\"%s\" " % (dsize, datastr)
    formatData = staticmethod(formatData)

    #@staticmethod
    def formatXid(xid, xidsize = None):
        if xid == None and xidsize != None:
            if xidsize > 0: raise Exception("Inconsistent XID size: xidsize=%d, xid=None" % xidsize)
            return ""
        if Utils.__isPrintable(xid):
            xidstr = Utils.__splitStr(xid)
        else:
            xidstr = Utils.__hexSplitStr(xid)
        if xidsize == None:
            xidsize = len(xid)
        elif xidsize != len(xid):
            raise Exception("Inconsistent XID size: xidsize=%d, xid(%d)=\"%s\"" % (xidsize, len(xid), xidstr))
        return "xid(%d)=\"%s\" " % (xidsize, xidstr)
    formatXid = staticmethod(formatXid)
    
    #@staticmethod
    def invStr(s):
        si = ""
        for i in range(0,len(s)):
            si += chr(~ord(s[i]) & 0xff)
        return si
    invStr = staticmethod(invStr)

    #@staticmethod
    def load(f, klass):
        args = Utils.__loadArgs(f, klass)
        subclass = klass.discriminate(args)
        result = subclass(*args) # create instance of record
        if subclass != klass:
            result.init(f, *Utils.__loadArgs(f, subclass))
        result.skip(f)
        return result;
    load = staticmethod(load)
    
    #@staticmethod
    def loadFileData(f, size, data):
        if size == 0:
            return (data, True)
        if data == None:
            loaded = 0
        else:
            loaded = len(data)
        foverflow = f.tell() + size - loaded > jfsize
        if foverflow:
            rsize = jfsize - f.tell()
        else:
            rsize = size - loaded
        bin = f.read(rsize)
        if data == None:
            data = unpack("%ds" % (rsize), bin)[0]
        else:
            data = data + unpack("%ds" % (rsize), bin)[0]
        return (data, not foverflow)
    loadFileData = staticmethod(loadFileData)

    #@staticmethod
    def remBytesInBlk(f, blkSize):
        foffs = f.tell()
        return Utils.sizeInBytesToBlk(foffs, blkSize) - foffs;
    remBytesInBlk = staticmethod(remBytesInBlk)

    #@staticmethod
    def sizeInBlks(size, blkSize):
        return int((size + blkSize - 1) / blkSize)
    sizeInBlks = staticmethod(sizeInBlks)

    #@staticmethod
    def sizeInBytesToBlk(size, blkSize):
        return Utils.sizeInBlks(size, blkSize) * blkSize
    sizeInBytesToBlk = staticmethod(sizeInBytesToBlk)

    #@staticmethod
    def __hexSplitStr(s, splitSize = 50):
        if len(s) <= splitSize:
            return Utils.__hexStr(s, 0, len(s))
#        if len(s) > splitSize + 25:
#            return Utils.__hexStr(s, 0, 10) + " ... " + Utils.__hexStr(s, 55, 65) + " ... " + Utils.__hexStr(s, len(s)-10, len(s))
        return Utils.__hexStr(s, 0, 10) + " ... " + Utils.__hexStr(s, len(s)-10, len(s))
    __hexSplitStr = staticmethod(__hexSplitStr)

    #@staticmethod
    def __hexStr(s, b, e):
        o = ""
        for i in range(b, e):
            if Utils.__isPrintable(s[i]):
                o += s[i]
            else:
                o += "\\%02x" % ord(s[i])
        return o
    __hexStr = staticmethod(__hexStr)

    #@staticmethod
    def __isPrintable(s):
        return s.strip(Utils.__printchars) == ""
    __isPrintable = staticmethod(__isPrintable)

    #@staticmethod
    def __loadArgs(f, klass):
        size = calcsize(klass.format)
        foffs = f.tell(),
        bin = f.read(size)
        if len(bin) != size: raise Exception("End of file")
        return foffs + unpack(klass.format, bin)
    __loadArgs = staticmethod(__loadArgs)

    #@staticmethod
    def __splitStr(s, splitSize = 50):
        if len(s) < splitSize:
            return s
        return s[:25] + " ... " + s[-25:]
    __splitStr = staticmethod(__splitStr)


#== class Warning =============================================================

class Warning(Exception):
    def __init__(self, err):
        Exception.__init__(self, err)


#== class Sizeable ============================================================

class Sizeable(object):

    def size(self):
        classes = [self.__class__]
        size = 0
        while classes:
            cls = classes.pop()
            if hasattr(cls, "format"):
                size += calcsize(cls.format)
            classes.extend(cls.__bases__)
        return size


#== class Hdr =================================================================

class Hdr(Sizeable):
 
    format = "=4sBBHQ"
    hdrVer = 1
    owi_mask = 0x01
    big_endian_flag = sys.byteorder == "big"

    def __init__(self, foffs, magic, ver, endn, flags, rid):
        self.foffs = foffs
        self.magic = magic
        self.ver = ver
        self.endn = endn
        self.flags = flags
        self.rid = long(rid)
        
    def __str__(self):
        if self.empty():
            return "0x%08x: <empty>" % (self.foffs)
        if self.magic[-1] == "x":
            return "0x%08x: [\"%s\"]" % (self.foffs, self.magic)
        if self.magic[-1] in ["a", "c", "d", "e", "f", "x"]:
            return "0x%08x: [\"%s\" v=%d e=%d f=0x%04x rid=0x%x]" % (self.foffs, self.magic, self.ver, self.endn, self.flags, self.rid)
        return "0x%08x: <error, unknown magic \"%s\" (possible overwrite boundary?)>" %  (self.foffs, self.magic)
    
    #@staticmethod
    def discriminate(args):
        return _CLASSES.get(args[1][-1], Hdr)
    discriminate = staticmethod(discriminate)

    def empty(self):
        return self.magic == "\x00"*4
    
    def encode(self):
        return pack(Hdr.format, self.magic, self.ver, self.endn, self.flags, self.rid)

    def owi(self):
        return self.flags & self.owi_mask != 0

    def skip(self, f):
        f.read(Utils.remBytesInBlk(f, dblkSize))

    def check(self):
        if self.empty() or self.magic[:3] != "RHM" or self.magic[3] not in ["a", "c", "d", "e", "f", "x"]:
            return True
        if self.magic[-1] != "x":
            if self.ver != self.hdrVer:
                raise Exception("%s: Invalid header version: found %d, expected %d." % (self, self.ver, self.hdrVer))
            if bool(self.endn) != self.big_endian_flag:
                if self.big_endian_flag: e = "big"
                else:                    e = "little"
                raise Exception("Endian mismatch: this platform is %s and does not match record encoding (0x%04x)" % (e, self.endn))
        return False
        

#== class FileHdr =============================================================

class FileHdr(Hdr):

    format = "=2H4x3Q"
        
    def __str__(self):
        return "%s fid=%d lid=%d fro=0x%08x t=%s" % (Hdr.__str__(self), self.fid, self.lid, self.fro, self.timestamp_str())
    
    def encode(self):
        return Hdr.encode(self) + pack(FileHdr.format, self.fid, self.lid, self.fro, self.time_sec, self.time_ns)

    def init(self, f, foffs, fid, lid, fro, time_sec, time_ns):
        self.fid = fid
        self.lid = lid
        self.fro = fro
        self.time_sec = time_sec
        self.time_ns = time_ns

    def skip(self, f):
        f.read(Utils.remBytesInBlk(f, sblkSize))

    def timestamp(self):
        return (self.time_sec, self.time_ns)

    def timestamp_str(self):
        ts = gmtime(self.time_sec)
        fstr = "%%a %%b %%d %%H:%%M:%%S.%09d %%Y" % (self.time_ns)
        return strftime(fstr, ts)


#== class DeqRec ==============================================================

class DeqRec(Hdr):

    format = "=QQ"

    def __str__(self):
        return "%s %sdrid=0x%x" % (Hdr.__str__(self), Utils.formatXid(self.xid, self.xidsize), self.deq_rid)

    def init(self, f, foffs, deq_rid, xidsize):
        self.deq_rid = deq_rid
        self.xidsize = xidsize
        self.xid = None
        self.deq_tail = None
        self.xid_complete = False
        self.tail_complete = False
        self.tail_bin = None
        self.tail_offs = 0
        self.load(f)
    
    def encode(self):
        d = Hdr.encode(self) + pack(DeqRec.format, self.deq_rid, self.xidsize)
        if self.xidsize > 0:
            fmt = "%ds" % (self.xidsize)
            d += pack(fmt, self.xid)
            d += self.deq_tail.encode()
        return d

    def load(self, f):
        if self.xidsize == 0:
            self.xid_complete = True
            self.tail_complete = True
        else:
            if not self.xid_complete:
                (self.xid, self.xid_complete) = Utils.loadFileData(f, self.xidsize, self.xid)
            if self.xid_complete and not self.tail_complete:
                ret = Utils.loadFileData(f, calcsize(RecTail.format), self.tail_bin)
                self.tail_bin = ret[0]
                if ret[1]:
                    self.deq_tail = RecTail(self.tail_offs, *unpack(RecTail.format, self.tail_bin))
                    if self.deq_tail.magic_inv != Utils.invStr(self.magic) or self.deq_tail.rid != self.rid:
                        raise Exception(" > %s *INVALID TAIL RECORD*" % self)
                    self.deq_tail.skip(f)
                self.tail_complete = ret[1]
        return self.complete()

    def complete(self):
        return self.xid_complete and self.tail_complete


#== class TxnRec ==============================================================

class TxnRec(Hdr):

    format = "=Q"

    def __str__(self):
        return "%s %s" % (Hdr.__str__(self), Utils.formatXid(self.xid, self.xidsize))

    def init(self, f, foffs, xidsize):
        self.xidsize = xidsize
        self.xid = None
        self.tx_tail = None
        self.xid_complete = False
        self.tail_complete = False
        self.tail_bin = None
        self.tail_offs = 0
        self.load(f)
    
    def encode(self):
         return Hdr.encode(self) + pack(TxnRec.format, self.xidsize) + pack("%ds" % self.xidsize, self.xid) + self.tx_tail.encode()

    def load(self, f):
        if not self.xid_complete:
            ret = Utils.loadFileData(f, self.xidsize, self.xid)
            self.xid = ret[0]
            self.xid_complete = ret[1]
        if self.xid_complete and not self.tail_complete:
            ret = Utils.loadFileData(f, calcsize(RecTail.format), self.tail_bin)
            self.tail_bin = ret[0]
            if ret[1]:
                self.tx_tail = RecTail(self.tail_offs, *unpack(RecTail.format, self.tail_bin))
                if self.tx_tail.magic_inv != Utils.invStr(self.magic) or self.tx_tail.rid != self.rid:
                    raise Exception(" > %s *INVALID TAIL RECORD*" % self)
                self.tx_tail.skip(f)
            self.tail_complete = ret[1]
        return self.complete()

    def complete(self):
        return self.xid_complete and self.tail_complete


#== class EnqRec ==============================================================

class EnqRec(Hdr):

    format = "=QQ"
    transient_mask = 0x10
    extern_mask = 0x20

    def __str__(self):
        return "%s %s%s %s %s" % (Hdr.__str__(self), Utils.formatXid(self.xid, self.xidsize), Utils.formatData(self.dsize, self.data), self.enq_tail, self.print_flags())
    
    def encode(self):
        d = Hdr.encode(self) + pack(EnqRec.format, self.xidsize, self.dsize)
        if self.xidsize > 0:
            d += pack("%ds" % self.xidsize, self.xid)
        if self.dsize > 0:
            d += pack("%ds" % self.dsize, self.data)
        if self.xidsize > 0 or self.dsize > 0:
            d += self.enq_tail.encode()
        return d

    def init(self, f, foffs, xidsize, dsize):
        self.xidsize = xidsize
        self.dsize = dsize
        self.transient = self.flags & self.transient_mask > 0
        self.extern = self.flags & self.extern_mask > 0
        self.xid = None
        self.data = None
        self.enq_tail = None
        self.xid_complete = False
        self.data_complete = False
        self.tail_complete = False
        self.tail_bin = None
        self.tail_offs = 0
        self.load(f)

    def load(self, f):
        if not self.xid_complete:
            ret = Utils.loadFileData(f, self.xidsize, self.xid)
            self.xid = ret[0]
            self.xid_complete = ret[1]
        if self.xid_complete and not self.data_complete:
            if self.extern:
                self.data_complete = True
            else:
                ret = Utils.loadFileData(f, self.dsize, self.data)
                self.data = ret[0]
                self.data_complete = ret[1]
        if self.data_complete and not self.tail_complete:
            ret = Utils.loadFileData(f, calcsize(RecTail.format), self.tail_bin)
            self.tail_bin = ret[0]
            if ret[1]:
                self.enq_tail = RecTail(self.tail_offs, *unpack(RecTail.format, self.tail_bin))
                if self.enq_tail.magic_inv != Utils.invStr(self.magic) or self.enq_tail.rid != self.rid:
                    raise Exception(" > %s *INVALID TAIL RECORD*" % self)
                self.enq_tail.skip(f)
            self.tail_complete = ret[1]
        return self.complete()

    def complete(self):
        return self.xid_complete and self.data_complete and self.tail_complete

    def print_flags(self):
        s = ""
        if self.transient:
            s = "*TRANSIENT"
        if self.extern:
            if len(s) > 0:
                s += ",EXTERNAL"
            else:
                s = "*EXTERNAL"
        if len(s) > 0:
            s += "*"
        return s


#== class RecTail =============================================================

class RecTail(Sizeable):

    format = "=4sQ"

    def __init__(self, foffs, magic_inv, rid):
        self.foffs = foffs
        self.magic_inv = magic_inv
        self.rid = long(rid)

    def __str__(self):
        magic = Utils.invStr(self.magic_inv)
        return "[\"%s\" rid=0x%x]" % (magic, self.rid)
    
    def encode(self):
        return pack(RecTail.format, self.magic_inv, self.rid)

    def skip(self, f):
        f.read(Utils.remBytesInBlk(f, dblkSize))


#== class EnqMap ==============================================================

class EnqMap(object):
    
    def __init__(self):
        self.__map = {}
    
    def __str__(self):
        return self.report(True, True)
        
    def add(self, fid, hdr, lock = False):
        if hdr.rid in self.__map.keys(): raise Exception("ERROR: Duplicate rid to EnqMap: rid=0x%x" % hdr.rid)
        self.__map[hdr.rid] = (fid, hdr, lock)
    
    def contains(self, rid):
        return rid in self.__map.keys()
    
    def delete(self, rid):
        if rid in self.__map.keys():
            if self.getLock(rid):
                raise Exception("ERROR: Deleting locked record from EnqMap: rid=0x%s" % rid)
            del self.__map[rid]
        else:
            raise Warning("ERROR: Deleting non-existent rid from EnqMap: rid=0x%x" % rid)
    
    def get(self, rid):
        if self.contains(rid): return self.__map[rid]
        return None
    
    def getFid(self, rid):
        if self.contains(rid): return self.__map[rid][0]
        return None
    
    def getHdr(self, rid):
        if self.contains(rid): return self.__map[rid][1]
        return None
    
    def getLock(self, rid):
        if self.contains(rid): return self.__map[rid][2]
        return None
    
    def getRecList(self):
        return self.__map.values()
    
    def lock(self, rid):
        if rid in self.__map.keys():
            tup = self.__map[rid]
            self.__map[rid] = (tup[0], tup[1], True)
        else:
            raise Warning("ERROR: Locking non-existent rid in EnqMap: rid=0x%x" % rid)
        
    def report(self, showStats, showRecords):
        if len(self.__map) == 0: return "No enqueued records found."
        s = "%d enqueued records found" % len(self.__map)
        if showRecords:
            s += ":"
            ridList = self.__map.keys()
            ridList.sort()
            for rid in ridList:
#            for f,r in self.__map.iteritems():
                r = self.__map[rid]
                if r[2]:
                    lockStr = " [LOCKED]"
                else:
                    lockStr = ""
                s += "\n  lfid=%d %s %s" % (r[0], r[1], lockStr)
        else:
            s += "."
        return s
    
    def rids(self):
        return self.__map.keys()
    
    def size(self):
        return len(self.__map)
    
    def unlock(self, rid):
        if rid in self.__map.keys():
            tup = self.__map[rid]
            if tup[2]:
                self.__map[rid] = (tup[0], tup[1], False)
            else:
                raise Exception("ERROR: Unlocking rid which is not locked in EnqMap: rid=0x%x" % rid)
        else:
            raise Exception("ERROR: Unlocking non-existent rid in EnqMap: rid=0x%x" % rid)


#== class TxnMap ==============================================================

class TxnMap(object):
    
    def __init__(self, emap):
        self.__emap = emap
        self.__map = {}
    
    def __str__(self):
        return self.report(True, True)
    
    def add(self, fid, hdr):
        if isinstance(hdr, DeqRec):
            self.__emap.lock(hdr.deq_rid)
        if hdr.xid in self.__map.keys():
            self.__map[hdr.xid].append((fid, hdr)) # append to existing list
        else:
            self.__map[hdr.xid] = [(fid, hdr)] # create new list
    
    def contains(self, xid):
        return xid in self.__map.keys()
    
    def delete(self, hdr):
        if hdr.magic[-1] == "c": return self.__commit(hdr.xid)
        if hdr.magic[-1] == "a": self.__abort(hdr.xid)
        else: raise Exception("ERROR: cannot delete from TxnMap using hdr type %s" % hdr.magic)
        
    def get(self, xid):
        if self.contains(xid): return self.__map[xid]
        
    def report(self, showStats, showRecords):
        if len(self.__map) == 0: return "No outstanding transactions found."
        s = "%d outstanding transactions found" % len(self.__map)
        if showRecords:
            s += ":"
            for x,t in self.__map.iteritems():
                s += "\n  xid=%s:" % Utils.formatXid(x)
                for i in t:
                    s += "\n   %s" % str(i[1])
        else:
            s += "."
        return s
    
    def size(self):
        return len(self.__map)
    
    def xids(self):
        return self.__map.keys()
    
    def __abort(self, xid):
        for fid, hdr in self.__map[xid]:
            if isinstance(hdr, DeqRec):
                self.__emap.unlock(hdr.rid)
        del self.__map[xid]
    
    def __commit(self, xid):
        mismatchList = []
        for fid, hdr in self.__map[xid]:
            if isinstance(hdr, EnqRec): 
                self.__emap.add(fid, hdr) # Transfer enq to emap
            else:
                if self.__emap.contains(hdr.deq_rid):
                    self.__emap.unlock(hdr.deq_rid)
                    self.__emap.delete(hdr.deq_rid)
                else:
                    mismatchList.append("0x%x" % hdr.deq_rid)
        del self.__map[xid]
        return mismatchList
                


#== class JrnlInfo ============================================================

class JrnlInfo(object):
    """
    This object reads and writes journal information files (<basename>.jinf). Methods are provided
    to read a file, query its properties and reset just those properties necessary for normalizing
    and resizing a journal.
    
    Normalizing: resetting the directory and/or base filename to different values. This is necessary
    if a set of journal files is copied from one location to another before being restored, as the
    value of the path in the file no longer matches the actual path.
    
    Resizing: If the journal geometry parameters (size and number of journal files) changes, then the
    .jinf file must reflect these changes, as this file is the source of information for journal
    recovery.
    
    NOTE: Size vs File size: There are methods which return the journal size and file size of the
    journal files.
    
    +-------------+--------------------/ /----------+
    | File header |           File data             |
    +-------------+--------------------/ /----------+
    |             |                                 |
    |             |<------------- Size ------------>|
    |<------------------ FileSize ----------------->|
    
    Size: The size of the data content of the journal, ie that part which stores the data records.
    
    File size: The actual disk size of the journal including data and the file header which precedes the
    data.
    
    The file header is fixed to 1 sblk, so  file size = jrnl size + sblk size.
    """
    
    def __init__(self, jdir, bfn = "JournalData"):
        self.__jdir = jdir
        self.__bfn = bfn
        self.__jinfDict = {}
        self.__read_jinf()
    
    def __str__(self):
        s = "Journal info file %s:\n" % os.path.join(self.__jdir, "%s.jinf" % self.__bfn)
        for k,v in self.__jinfDict.iteritems():
            s += "  %s = %s\n" % (k, v)
        return s
    
    def normalize(self, jdir = None, baseFilename = None):
        if jdir == None:
            self.__jinfDict["directory"] = self.__jdir
        else:
            self.__jdir = jdir
            self.__jinfDict["directory"] = jdir
        if baseFilename != None:
            self.__bfn = baseFilename
            self.__jinfDict["base_filename"] = baseFilename
    
    def resize(self, njFiles = None, jfSize = None):
        if njFiles != None:
            self.__jinfDict["number_jrnl_files"] = njFiles
        if jfSize != None:
            self.__jinfDict["jrnl_file_size_sblks"] = jfSize * dblkSize

    def write(self, jdir = None, baseFilename = None):
        self.normalize(jdir, baseFilename)
        if not os.path.exists(self.getJrnlDir()): os.makedirs(self.getJrnlDir())
        wdir = os.path.join(self.getJrnlDir(), "%s.jinf" % self.getJrnlBaseName())
        f = open(os.path.join(self.getJrnlDir(), "%s.jinf" % self.getJrnlBaseName()), "w")
        f.write("<?xml version=\"1.0\" ?>\n")
        f.write("<jrnl>\n")
        f.write("  <journal_version value=\"%d\" />\n" % self.getJrnlVersion())
        f.write("  <journal_id>\n")
        f.write("    <id_string value=\"%s\" />\n" % self.getJrnlId())
        f.write("    <directory value=\"%s\" />\n" % self.getJrnlDir())
        f.write("    <base_filename value=\"%s\" />\n" % self.getJrnlBaseName())
        f.write("  </journal_id>\n")
        f.write("  <creation_time>\n")
        f.write("    <seconds value=\"%d\" />\n" % self.getCreationTime()[0])
        f.write("    <nanoseconds value=\"%d\" />\n" % self.getCreationTime()[1])
        f.write("    <string value=\"%s\" />\n" % self.getCreationTimeStr())
        f.write("  </creation_time>\n")
        f.write("  <journal_file_geometry>\n")
        f.write("    <number_jrnl_files value=\"%d\" />\n" % self.getNumJrnlFiles())
        f.write("    <auto_expand value=\"%s\" />\n" % str.lower(str(self.getAutoExpand())))
        f.write("    <jrnl_file_size_sblks value=\"%d\" />\n" % self.getJrnlSizeSblks())
        f.write("    <JRNL_SBLK_SIZE value=\"%d\" />\n" % self.getJrnlSblkSize())
        f.write("    <JRNL_DBLK_SIZE value=\"%d\" />\n" % self.getJrnlDblkSize())
        f.write("  </journal_file_geometry>\n")
        f.write("  <cache_geometry>\n")
        f.write("    <wcache_pgsize_sblks value=\"%d\" />\n" % self.getWriteBufferPageSizeSblks())
        f.write("    <wcache_num_pages value=\"%d\" />\n" % self.getNumWriteBufferPages())
        f.write("    <JRNL_RMGR_PAGE_SIZE value=\"%d\" />\n" % self.getReadBufferPageSizeSblks())
        f.write("    <JRNL_RMGR_PAGES value=\"%d\" />\n" % self.getNumReadBufferPages())
        f.write("  </cache_geometry>\n")
        f.write("</jrnl>\n")
        f.close()
    
    # Journal ID
    
    def getJrnlVersion(self):
        return self.__jinfDict["journal_version"]
    
    def getJrnlId(self):
        return self.__jinfDict["id_string"]
    
    def getCurrentJnrlDir(self):
        return self.__jdir
    
    def getJrnlDir(self):
        return self.__jinfDict["directory"]
    
    def getJrnlBaseName(self):
        return self.__jinfDict["base_filename"]
    
    # Journal creation time
    
    def getCreationTime(self):
        return (self.__jinfDict["seconds"], self.__jinfDict["nanoseconds"])
    
    def getCreationTimeStr(self):
        return self.__jinfDict["string"]
    
    # Files and geometry
    
    def getNumJrnlFiles(self):
        return self.__jinfDict["number_jrnl_files"]
    
    def getAutoExpand(self):
        return self.__jinfDict["auto_expand"]
    
    def getJrnlSblkSize(self):
        return self.__jinfDict["JRNL_SBLK_SIZE"]
    
    def getJrnlDblkSize(self):
        return self.__jinfDict["JRNL_DBLK_SIZE"]
    
    def getJrnlSizeSblks(self):
        return self.__jinfDict["jrnl_file_size_sblks"]
    
    def getJrnlSizeDblks(self):
        return self.getJrnlSizeSblks() * self.getJrnlSblkSize()
    
    def getJrnlSizeBytes(self):
        return self.getJrnlSizeDblks() * self.getJrnlDblkSize()
    
    def getJrnlFileSizeSblks(self):
        return self.getJrnlSizeSblks() + 1
    
    def getJrnlFileSizeDblks(self):
        return self.getJrnlFileSizeSblks() * self.getJrnlSblkSize()
    
    def getJrnlFileSizeBytes(self):
        return self.getJrnlFileSizeDblks() * self.getJrnlDblkSize()
    
    def getTotalJrnlFileCapacitySblks(self):
        return self.getNumJrnlFiles() * self.getJrnlSizeBytes()
    
    def getTotalJrnlFileCapacityDblks(self):
        return self.getNumJrnlFiles() * self.getJrnlSizeDblks()
    
    def getTotalJrnlFileCapacityBytes(self):
        return self.getNumJrnlFiles() * self.getJrnlSizeBytes()
    
    # Read and write buffers
    
    def getWriteBufferPageSizeSblks(self):
        return self.__jinfDict["wcache_pgsize_sblks"]
    
    def getWriteBufferPageSizeDblks(self):
        return self.getWriteBufferPageSizeSblks() * self.getJrnlSblkSize()
    
    def getWriteBufferPageSizeBytes(self):
        return self.getWriteBufferPageSizeDblks() * self.getJrnlDblkSize()
    
    def getNumWriteBufferPages(self):
        return self.__jinfDict["wcache_num_pages"]
    
    def getReadBufferPageSizeSblks(self):
        return self.__jinfDict["JRNL_RMGR_PAGE_SIZE"]
    
    def getReadBufferPageSizeDblks(self):
        return self.getReadBufferPageSizeSblks * self.getJrnlSblkSize()
    
    def getReadBufferPageSizeBytes(self):
        return self.getReadBufferPageSizeDblks * self.getJrnlDblkSize()
    
    def getNumReadBufferPages(self):
        return self.__jinfDict["JRNL_RMGR_PAGES"]
    
    def __read_jinf(self):
        f = open(os.path.join(self.__jdir, "%s.jinf" % self.__bfn), "r")
        p = xml.parsers.expat.ParserCreate()
        p.StartElementHandler = self.__handleXmlStartElement
        p.CharacterDataHandler = self.__handleXmlCharData
        p.EndElementHandler = self.__handleXmlEndElement
        p.ParseFile(f)
        f.close()

    def __handleXmlStartElement(self, name, attrs):
        # bool values
        if name == "auto_expand":
            self.__jinfDict[name] = attrs["value"] == "true"
        # long values
        elif name == "seconds" or \
             name == "nanoseconds":
            self.__jinfDict[name] = long(attrs["value"])
        # int values
        elif name == "journal_version" or \
             name == "number_jrnl_files" or \
             name == "jrnl_file_size_sblks" or \
             name == "JRNL_SBLK_SIZE" or \
             name == "JRNL_DBLK_SIZE" or \
             name == "wcache_pgsize_sblks" or \
             name == "wcache_num_pages" or \
             name == "JRNL_RMGR_PAGE_SIZE" or \
             name == "JRNL_RMGR_PAGES":
            self.__jinfDict[name] = int(attrs["value"])
        # strings
        elif "value" in attrs:
            self.__jinfDict[name] = attrs["value"]

    def __handleXmlCharData(self, data): pass

    def __handleXmlEndElement(self, name): pass

#== class JrnlAnalyzer ========================================================

class JrnlAnalyzer(object):
    """
    This class analyzes a set of journal files and determines which is the last to be written
    (the newest file),  and hence which should be the first to be read for recovery (the oldest
    file).
    
    The analysis is performed on construction; the contents of the JrnlInfo object passed provide
    the recovery details.
    """

    def __init__(self, jinf):
        self.__oldest = None
        self.__jinf = jinf
        self.__flist = self._analyze()
                
    def __str__(self):
        s = "Journal files analyzed in directory %s (* = earliest full):\n" % self.__jinf.getCurrentJnrlDir()
        if self.isEmpty():
            s += "  <All journal files are empty>\n"
        else:
            for tup in self.__flist:
                o = " "
                if tup[0] == self.__oldest[0]: o = "*"
                s += "  %s %s: owi=%-5s rid=0x%x, fro=0x%x ts=%s\n" % (o, os.path.basename(tup[1]), tup[2], tup[3], tup[4], tup[5])
            for i in range(self.__flist[-1][0] + 1, self.__jinf.getNumJrnlFiles()):
                s += "    %s.%04d.jdat: <empty>\n" % (self.__jinf.getJrnlBaseName(), i) 
        return s
        
    # Analysis
    
    def getOldestFile(self):
        return self.__oldest

    def getOldestFileIndex(self):
        if self.isEmpty(): return None
        return self.__oldest[0]

    def isEmpty(self):
        return len(self.__flist) == 0
    
    def _analyze(self):
        fname = ""
        fnum = -1
        rid = -1
        fro = -1
        tss = ""
        owi_found = False
        flist = []
        for i in range(0, self.__jinf.getNumJrnlFiles()):
            jfn = os.path.join(self.__jinf.getCurrentJnrlDir(), "%s.%04x.jdat" % (self.__jinf.getJrnlBaseName(), i))
            f = open(jfn)
            fhdr = Utils.load(f, Hdr)
            if fhdr.empty(): break
            this_tup = (i, jfn, fhdr.owi(), fhdr.rid, fhdr.fro, fhdr.timestamp_str())
            flist.append(this_tup)
            if i == 0:
                init_owi = fhdr.owi()
                self.__oldest = this_tup
            elif fhdr.owi() != init_owi and not owi_found:
                self.__oldest = this_tup
                owi_found = True
        return flist
        

#== class JrnlReader ====================================================

class JrnlReader(object):
    """
    This class contains an Enqueue Map (emap), a transaction map (tmap) and a transaction
    object list (txnObjList) which are populated by reading the journals from the oldest
    to the newest and analyzing each record. The JrnlInfo and JrnlAnalyzer
    objects supplied on construction provide the information used for the recovery.
    
    The analysis is performed on construction.
    """
    
    def __init__(self, ji, jra, qflag = False, rflag = False, vflag = False):
        self._ji = ji
        self._jra = jra
        self._qflag = qflag
        self._rflag = rflag
        self._vflag = vflag
        
        # test callback functions for CSV tests
        self._csvStoreChk = None
        self._csvStartCb = None
        self._csvEnqCb = None
        self._csvDeqCb = None
        self._csvTxnCb = None
        self._csvEndCb = None
        
        self._emap = EnqMap()
        self._tmap = TxnMap(self._emap)
        self._txnObjList = {}
        
        self._file = None
        self._fileHdr = None
        self._fileNum = None
        self._firstRecFlag = None
        self._fro = None
        self._lastFileFlag = None
        self._startFileNum = None
        self._warning = []
        
        self._abortCnt = 0
        self._commitCnt = 0
        self._msgCnt = 0
        self._recCnt = 0
        self._txnMsgCnt = 0
    
    def __str__(self):
        return self.report(True, self._rflag)

    def abortCnt(self): return self._abortCnt

    def commitCnt(self): return self._commitCnt
    
    def emap(self): return self._emap
    
    def msgCnt(self): return self._msgCnt
    
    def recCnt(self): return self._recCnt
    
    def report(self, showStats = True, showRecords = False):
        s = self._emap.report(showStats, showRecords) + "\n" + self._tmap.report(showStats, showRecords)
        #TODO - print size analysis here
        return s
    
    def run(self):
        if self._csvStartCb != None and self._csvStartCb(self._csvStoreChk): return
        if self._jra.isEmpty(): return
        stop = self._advanceJrnlFile(*self._jra.getOldestFile())
        while not stop and not self._getNextRecord(): pass
        if self._csvEndCb != None and self._csvEndCb(self._csvStoreChk): return
        if not self._qflag: print
    
    def setCallbacks(self, csvStoreChk, csvStartCb = None, csvEnqCb = None, csvDeqCb = None, csvTxnCb = None, csvEndCb = None):
        self._csvStoreChk = csvStoreChk
        self._csvStartCb = csvStartCb
        self._csvEnqCb = csvEnqCb
        self._csvDeqCb = csvDeqCb
        self._csvTxnCb = csvTxnCb
        self._csvEndCb = csvEndCb
    
    def tmap(self): return self._tmap
    
    def txnMsgCnt(self): return self._txnMsgCnt
    
    def txnObjList(self): return self._txnObjList
    
    def _advanceJrnlFile(self, *oldestFileInfo):
        froSeekFlag = False
        if len(oldestFileInfo) > 0:
            self._startFileNum = self._fileNum = oldestFileInfo[0]
            self._fro = oldestFileInfo[4]
            froSeekFlag = True # jump to fro to start reading
            if not self._qflag and not self._rflag:
                if self._vflag: print "Recovering journals..."
                else: print "Recovering journals",
        if self._file != None and self._fileFull():
            self._file.close()
            self._fileNum = self._incrFileNum()
            if self._fileNum == self._startFileNum:
                return True
            if self._startFileNum == 0:
                self._lastFileFlag = self._fileNum == self._ji.getNumJrnlFiles() - 1
            else:
                self._lastFileFlag = self._fileNum == self._startFileNum - 1
        if self._fileNum < 0 or self._fileNum >= self._ji.getNumJrnlFiles():
            raise Exception("Bad file number %d" % self._fileNum)
        jfn = os.path.join(self._ji.getCurrentJnrlDir(), "%s.%04x.jdat" % (self._ji.getJrnlBaseName(), self._fileNum))
        self._file = open(jfn)
        self._fileHdr = Utils.load(self._file, Hdr)
        if froSeekFlag and self._file.tell() != self._fro:
            self._file.seek(self._fro)
        self._firstRecFlag = True
        if not self._qflag:
            if self._rflag: print jfn, ": ", self._fileHdr
            elif self._vflag: print "* Reading %s" % jfn
            else:
                print ".",
                sys.stdout.flush()
        return False

    def _checkOwi(self, hdr):
        return self._fileHdrOwi == hdr.owi()
    
    def _fileFull(self):
        return self._file.tell() >= self._ji.getJrnlFileSizeBytes()
    
    def _getNextRecord(self, *oldestFileInfo):
        if self._fileFull():
            if self._advanceJrnlFile(): return True
        try: hdr = Utils.load(self._file, Hdr)
        except: return True
        if hdr.empty(): return True
        if hdr.check(): return True
        self._recCnt += 1
        self._fileHdrOwi = self._fileHdr.owi()
        if self._firstRecFlag:
            if self._fileHdr.fro != hdr.foffs:
                raise Exception("File header first record offset mismatch: fro=0x%x; rec_offs=0x%x" % (self._fileHdr.fro, hdr.foffs))
            else:
                if self._rflag: print " * fro ok: 0x%x" % self._fileHdr.fro
                self._firstRecFlag = False
        stop = False
        if   isinstance(hdr, EnqRec):
            stop = self._handleEnqRec(hdr)
        elif isinstance(hdr, DeqRec):
            stop = self._handleDeqRec(hdr)
        elif isinstance(hdr, TxnRec):
            stop = self._handleTxnRec(hdr)
        wstr = ""
        for w in self._warning:
            wstr += " (%s)" % w
        if self._rflag: print " > %s  %s" % (hdr, wstr)
        self._warning = []
        return stop
     
    def _handleDeqRec(self, hdr):
        if self._loadRec(hdr): return True
        
        # Check OWI flag
        if not self._checkOwi(hdr):
            self._warning.append("WARNING: OWI mismatch - could be overwrite boundary.")
            return True
        # Test hook
        if self._csvDeqCb != None and self._csvDeqCb(self._csvStoreChk, hdr): return True
        
        try:
            if hdr.xid == None:
                self._emap.delete(hdr.deq_rid)
            else:
                self._tmap.add(self._fileHdr.fid, hdr)
        except Warning, w: self._warning.append(str(w))
        return False
    
    def _handleEnqRec(self, hdr):
        if self._loadRec(hdr): return True
        
        # Check extern flag
        if hdr.extern and hdr.data != None: raise Exception("Message data found on external record: hdr=%s" % hdr)
        # Check OWI flag
        if not self._checkOwi(hdr):
            self._warning.append("WARNING: OWI mismatch - could be overwrite boundary.")
            return True
        # Test hook
        if self._csvEnqCb != None and self._csvEnqCb(self._csvStoreChk, hdr): return True
        
        if hdr.xid == None:
            self._emap.add(self._fileHdr.fid, hdr)
        else:
            self._txnMsgCnt += 1
            self._tmap.add(self._fileHdr.fid, hdr)
        self._msgCnt += 1
        return False
    
    def _handleTxnRec(self, hdr):
        if self._loadRec(hdr): return True
        
        # Check OWI flag
        if not self._checkOwi(hdr):
            self._warning.append("WARNING: OWI mismatch - could be overwrite boundary.")
            return True
        # Test hook
        if self._csvTxnCb != None and self._csvTxnCb(self._csvStoreChk, hdr): return True
               
        if hdr.magic[-1] == "a": self._abortCnt += 1
        else: self._commitCnt += 1
        
        if self._tmap.contains(hdr.xid):
            mismatchedRids = self._tmap.delete(hdr)
            if mismatchedRids != None and len(mismatchedRids) > 0:
                self._warning.append("WARNING: transactional dequeues not found in enqueue map; rids=%s" % mismatchedRids)
        else:
            self._warning.append("WARNING: %s not found in transaction map" % Utils.formatXid(hdr.xid))
        if hdr.magic[-1] == "c": # commits only
            self._txnObjList[hdr.xid] = hdr
        return False

    def _incrFileNum(self):
        self._fileNum += 1
        if self._fileNum >= self._ji.getNumJrnlFiles():
            self._fileNum = 0;
        return self._fileNum
    
    def _loadRec(self, hdr):
        while not hdr.complete():
            if self._advanceJrnlFile(): return True
            hdr.load(self._file)
        return False
        

#==============================================================================

_CLASSES = {
    "a": TxnRec,
    "c": TxnRec,
    "d": DeqRec,
    "e": EnqRec,
    "f": FileHdr
}

if __name__ == "__main__":
    print "This is a library, and cannot be executed."
